"""Define a client to interact with the Ambient Weather APIs."""
# pylint: disable=unused-import
import logging
from typing import Dict, List, Union  # noqa

from aiohttp import ClientSession

from .api import API
from .websocket import Websocket

_LOGGER = logging.getLogger(__name__)

DEFAULT_API_VERSION = 1


class Client:  # pylint: disable=too-few-public-methods
    """Define the client."""

    def __init__(
            self,
            api_key: str,
            application_key: str,
            session: ClientSession,
            *,
            api_version: int = DEFAULT_API_VERSION) -> None:
        """Initialize."""
        self.api = API(application_key, api_key, api_version, session)
        self.websocket = Websocket(application_key, api_key, api_version)



"""Define an object to interact with the REST API."""
import asyncio
from datetime import datetime

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientError

from .errors import RequestError

REST_API_BASE = 'https://api.ambientweather.net'
DEFAULT_LIMIT = 288


class API:
    """Define to handler."""

    def __init__(
            self, application_key: str, api_key: str, api_version: int,
            session: ClientSession) -> None:
        """Initialize."""
        self._api_key = api_key
        self._api_version = api_version
        self._application_key = application_key
        self._session = session

    async def _request(
            self, method: str, endpoint: str, *, params: dict = None) -> list:
        """Make a request against air-matters.com."""
        # In order to deal with Ambient's fairly aggressive rate limiting, we
        # pause for a second before continuing in case any requests came before
        # this.
        # https://ambientweather.docs.apiary.io/#introduction/rate-limiting
        await asyncio.sleep(1)

        url = '{0}/v{1}/{2}'.format(REST_API_BASE, self._api_version, endpoint)

        if not params:
            params = {}
        params.update({
            "apiKey": self._api_key,
            "applicationKey": self._application_key
        })

        async with self._session.request(method, url, params=params) as resp:
            try:
                resp.raise_for_status()
                return await resp.json(content_type=None)
            except ClientError as err:
                raise RequestError(
                    "Error requesting data from {0}: {1}".format(url, err))

    async def get_devices(self) -> list:
        """Get all devices associated with an API key."""
        return await self._request("get", "devices")

    async def get_device_details(
            self,
            mac_address: str,
            *,
            end_date: datetime = None,
            limit: int = DEFAULT_LIMIT) -> list:
        """Get details of a device by MAC address."""
        params = {"limit": limit}
        if end_date:
            params["endDate"] = end_date.isoformat()  # type: ignore

        return await self._request(
            "get", "devices/{0}".format(mac_address), params=params)



"""Define fixtures for testing the REST API."""
import pytest


@pytest.fixture()
def device_details_json():
    """Define a response to /v1/devices/MAC_ADDRESS."""
    return [
        {
            "dateutc": 1547094300000,
            "winddir": 344,
            "windspeedmph": 1.6,
            "windgustmph": 2.2,
            "maxdailygust": 3.4,
            "tempf": 34,
            "hourlyrainin": 0,
            "eventrainin": 0,
            "dailyrainin": 0,
            "weeklyrainin": 0,
            "monthlyrainin": 0,
            "totalrainin": 0,
            "baromrelin": 30.38,
            "baromabsin": 24.89,
            "humidity": 49,
            "tempinf": 69.6,
            "humidityin": 30,
            "uv": 0,
            "solarradiation": 0,
            "feelsLike": 34,
            "dewPoint": 16.87,
            "date": "2019-01-10T04:25:00.000Z",
        },
        {
            "dateutc": 1547094000000,
            "winddir": 344,
            "windspeedmph": 0,
            "windgustmph": 0,
            "maxdailygust": 3.4,
            "tempf": 34,
            "hourlyrainin": 0,
            "eventrainin": 0,
            "dailyrainin": 0,
            "weeklyrainin": 0,
            "monthlyrainin": 0,
            "totalrainin": 0,
            "baromrelin": 30.38,
            "baromabsin": 24.89,
            "humidity": 50,
            "tempinf": 69.4,
            "humidityin": 29,
            "uv": 0,
            "solarradiation": 0,
            "feelsLike": 34,
            "dewPoint": 17.34,
            "date": "2019-01-10T04:20:00.000Z",
        },
    ]


@pytest.fixture()
def devices_json():
    """Define a response to /v1/devices."""
    return [
        {
            "macAddress": "84:F3:EB:21:90:C4",
            "lastData": {
                "dateutc": 1546889640000,
                "baromrelin": 30.09,
                "baromabsin": 24.61,
                "tempinf": 68.9,
                "humidityin": 30,
                "date": "2019-01-07T19:34:00.000Z",
            },
            "info": {
                "name": "Home",
                "location": "Home"
            },
        },
        {
            "macAddress": "84:F3:EB:21:90:C4",
            "lastData": {
                "dateutc": 1546889640000,
                "baromrelin": 30.09,
                "baromabsin": 24.61,
                "tempinf": 68.9,
                "humidityin": 30,
                "date": "2019-01-06T19:34:00.000Z",
            },
            "info": {
                "name": "Home",
                "location": "Home"
            },
        },
    ]



"""Define an object to interact with the Websocket API."""
# pylint: disable=redefined-builtin,unused-argument
from typing import Awaitable, Callable, Union

from socketio import AsyncClient
from socketio.exceptions import ConnectionError, SocketIOError

from .errors import WebsocketError

WEBSOCKET_API_BASE = 'https://dash2.ambientweather.net'


class Websocket:
    """Define to handler."""

    def __init__(
            self, application_key: str, api_key: str,
            api_version: int) -> None:
        """Initialize."""
        self._api_key = api_key
        self._api_version = api_version
        self._application_key = application_key
        self._sio = AsyncClient()
        self._user_connect_handler = None

    async def _init_connection(self) -> None:
        """Perform automatic initialization upon connecting."""
        await self._sio.emit('subscribe', {'apiKeys': [self._api_key]})

        if self._user_connect_handler:
            self._user_connect_handler()

    def on_connect(self, target: Union[Awaitable, Callable]) -> None:
        """Define a method/coroutine to be called when connecting."""
        self._user_connect_handler = target  # type: ignore

    def on_data(self, target: Union[Awaitable, Callable]) -> None:
        """Define a method/coroutine to be called when data is received."""
        self._sio.on('data', target)

    def on_disconnect(self, target: Union[Awaitable, Callable]) -> None:
        """Define a method/coroutine to be called when disconnecting."""
        self._sio.on('disconnect', target)

    def on_subscribed(self, target: Union[Awaitable, Callable]) -> None:
        """Define a method/coroutine to be called when subscribed."""
        self._sio.on('subscribed', target)

    async def connect(self) -> None:
        """Connect to the socket."""
        try:
            self._sio.on('connect', self._init_connection)
            await self._sio.connect(
                '{0}/?api={1}&applicationKey={2}'.format(
                    WEBSOCKET_API_BASE, self._api_version,
                    self._application_key),
                transports=['websocket'])
        except (ConnectionError, SocketIOError) as err:
            raise WebsocketError(err) from None

    async def disconnect(self) -> None:
        """Disconnect from the socket."""
        await self._sio.disconnect()
